/*
// cascade.cc
// 5-2002  - 8- 2002
// olaf booij 
// obooij@science.uva.nl
*/
/*
// output-files:
// in blog1.data:
// per output-epoch
// aantal goede tests (van de 24),  welke rule wordt gebruikt, aantal confl_wei fouten
//
// in blogall1.data:
// per output-epoch:
// over alle patronen:
// het aanal errors per probleem-groep
//
// in blogb1.data:
// output-epoch-nr waarin een hidden werd toegevoegd
//
// in BalanceLog.data:
// per testsessie:
// het aantal epochs en het aantal hiddens toegevoegd
*/

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <ctime>
#include "weightlist.cc"
#include "pattern.cc" 

using namespace std;

class Neural_networkC {};

class Cascade_netC : public Neural_networkC {
public:
  int nrofepochs;
  Cascade_netC (int nroi, int nroo) {
    int i, j;
    nrofinputs = nroi;
    nrofoutputs = nroo;
    nrofhiddens = 0;
    nrofcandidates = 0;
    nrofepochs = 0;
    inputunit = new Unit_array(nrofinputs);
    for(j=0;j<nrofinputs;j++)
      inputunit->set(j, new InputUnitC());
    outputunit = new Unit_array(nrofoutputs);
    for(i=0;i<nrofoutputs;i++)
      outputunit->set(i, new OutputUnitC());
    hiddenunit = new Unit_array(nrofhiddens);
    for(i=0; i < nrofoutputs; i++)
      for(j=0; j < nrofinputs; j++)
        outputunit->get(i)->add_connection(inputunit->get(j));
    candidateunit = NULL;
  }
  ~Cascade_netC() {
    int i, j;
    for(j=0;j<nrofinputs;j++)
      delete inputunit->get(j);
    delete inputunit;
    for(i=0;i<nrofoutputs;i++)
      delete outputunit->get(i);
    delete outputunit;
    for(i=0;i<nrofhiddens;i++)
      delete hiddenunit->get(i);
    delete hiddenunit;
    if (candidateunit) {
      for(i=0;i<nrofcandidates;i++)
        delete candidateunit->get(i);
      delete candidateunit;
    }
  }
  void infoprint() {
    int i;
    cout << "netwerk-inf: " 
    //     << "Aantal input-knopen: " << nrofinputs << endl
    //     << "Aantal output-knopen: " << nrofoutputs << endl
         << "Aantal hidden-knopen: " << nrofhiddens << endl;
    //     << "Aantal kandidaat-knopen: " << nrofcandidates << endl;
    //cout << "connecties van output-knopen: " << endl;
    for(i=0; i<nrofhiddens; i++)
      hiddenunit->get(i)->show();
    for(i=0; i<nrofoutputs; i++)
      outputunit->get(i)->show();
  }
  void output_phase(Pattern_setC* patternset, Pattern_setC* testset, Pattern_setC* testset_all) {
    double error = 1.0,
           old_error = 0.0;
    int error_all[6];
    int output_patience,
        testfout,
        rule,
        confl_wei_error,
        i;
    const int nrofe_atbegin = nrofepochs; 
    fstream outputfile2("blog1.data", ios::app);
    fstream outputfileall("blogall1.data", ios::app);
    for (output_patience=0; ((output_patience < 8) && ((nrofepochs - nrofe_atbegin) < 200)) && nrofepochs < 300; output_patience++) {
      //infoprint();
      nrofepochs++;
      trainoutput_set(patternset, &error);
      if (fabs(error - old_error) > (old_error * 0.015)) {
        output_patience = 0;
        old_error = error;
      }
      if (nrofepochs % 10 == 0)
        cout << " de echte fout is :" << error << endl;
      testfout = test_set(testset);
      test_set_new(testset_all, error_all);
      rule = rule_diagnose(testset, &confl_wei_error);
      cout << "de testset-fout is:" << testfout;
      cout << " rule: " << rule ;
      cout << " confl_weight_error: " << confl_wei_error << endl;
      outputfile2 << (24 - testfout) << "  " << rule << "  " << confl_wei_error << endl;
      for (i=0; i < 6; i++)
        outputfileall << "  " << error_all[i];
      outputfileall << endl;
      patternset->expand();
    }
  }
  void input_phase(Pattern_setC* patternset) {
    int input_patience;
    double best_candidate_value,
           old_best_candidate_value = 0.0;
    int best_candidate_index = -1;
    int nrofinputepochs = 0; 
    newcandidates();
    for (input_patience=0; (input_patience < 8) && (nrofinputepochs < 50); input_patience++) {
      //infoprint();
      nrofinputepochs++;
      traininput_set(patternset, &best_candidate_value, &best_candidate_index);
      if (fabs(best_candidate_value - old_best_candidate_value) > (old_best_candidate_value * 0.03)) {
        input_patience = 0;
        old_best_candidate_value = best_candidate_value;
      }
      if (nrofinputepochs % 10 == 0)
        cout << "input_phase: de hoogste waarde is: " << best_candidate_value << " van kandidaat: " << best_candidate_index << endl;
    }
    movecandidate(best_candidate_index);
    //nrofepochs += nrofinputepochs;
    cout << "aantal inputepochs is: " <<  nrofinputepochs << endl;
  }
  int test_set(Pattern_setC* testingset) {
    int k;
    int error = 0;
    for (k=0; k < testingset->nrofpatterns; k++)
      error += test_pattern(testingset->pattern[k]);
   return error;
  }
  void test_set_new(Pattern_setC* testingset, int* error) {
    int k;
    for (k=0; k < 6; k++)
      error[k] = 0;
    for (k=0; k < 25; k++)
      error[0] += test_pattern(testingset->pattern[k]);
    for (k=25; k < 125; k++)
      error[1] += test_pattern(testingset->pattern[k]);
    for (k=125; k < 225; k++)
      error[2] += test_pattern(testingset->pattern[k]);
    for (k=225; k < 313; k++)
      error[3] += test_pattern(testingset->pattern[k]);
    for (k=313; k < 401; k++)
      error[4] += test_pattern(testingset->pattern[k]);
    for (k=401; k < 425; k++)
      error[5] += test_pattern(testingset->pattern[k]);
  }
  int rule_diagnose(Pattern_setC* testingset, int* confl_wei_error) {
    int k;
    int error[24];
    if (testingset->nrofpatterns != 24) 
      cerr << "rule_diagnose: error: testingset is no good" << endl;
    for (k=0; k < testingset->nrofpatterns; k++)
      error[k] = test_pattern(testingset->pattern[k]);
    // conflict_weight_score:
    *confl_wei_error = count_array(error, 12, 16);
    // rule 4:
    if (count_array(error, 0, 24) >= 20)
      return(4);
    // rule 2:
    if ((count_array(error, 0, 16) >= 13) && (count_array(error, 16, 24) < 3))
      return(2);
    // rule 3:
    if ((count_array(error, 0, 11) >= 10) && (count_array(error, 12, 24) < 10))
      return(3);
    // rule 1:
    if (((count_array(error, 0, 8) + count_array(error, 12, 16)) >= 10) && ((count_array(error, 8, 12) + count_array(error, 16, 24)) < 3))
      return(1);
    return(0);
  }
  int count_array(int *array, int begin, int end) {
    int i,
        counter = 0;
    for (i = begin; i < end; i++)
      if (!array[i]) 
        counter++;
    return(counter);
  }
private: 
  int nrofinputs,
      nrofoutputs,
      nrofhiddens,
      nrofcandidates;
  Unit_array *inputunit,
             *outputunit,
             *hiddenunit,
             *candidateunit;
  void newcandidates() {
    int i,j;
    nrofcandidates = 8;
    candidateunit = new Unit_array(nrofcandidates);
    for (i=0; i < nrofcandidates; i++) {
      candidateunit->set(i, new HiddenUnitC());
      for(j=0; j < nrofinputs; j++)
        candidateunit->get(i)->add_connection(inputunit->get(j));
      for(j=0; j < nrofhiddens; j++)
        candidateunit->get(i)->add_connection(hiddenunit->get(j));
    }
  }
  void movecandidate(int winnercandidate) {
    int i,j;
    hiddenunit->add();
    nrofhiddens++;
    hiddenunit->set(nrofhiddens-1, candidateunit->get(winnercandidate));
    for(i=0; i<nrofoutputs; i++)
      outputunit->get(i)->add_connection(hiddenunit->get(nrofhiddens-1));
    for(i=0;i<nrofcandidates;i++)
      if (i != winnercandidate)
        delete candidateunit->get(i);
    delete candidateunit;
    candidateunit = NULL;
    nrofcandidates = 0;
  }
  void clamp(PatternC* pattern) {
    int i,j;
    for (j=0; j < nrofinputs; j++)
      inputunit->get(j)->clamp(pattern->input[j]);
    for (i=0; i < nrofoutputs; i++)
      outputunit->get(i)->set_target(pattern->target[i]);
  } 
  void propagate() {
    int i;
    for (i=0; i < nrofhiddens; i++)
      hiddenunit->get(i)->propagate();  // hier is de volgorde van belang
    for (i=0; i < nrofoutputs; i++)
      outputunit->get(i)->propagate();
    for (i=0; i < nrofcandidates; i++)
      candidateunit->get(i)->propagate();
  }
  void trainoutput_pattern(PatternC* trainingpattern, double *pattern_error) {
    int i;
    *pattern_error = 0.0;
    clamp(trainingpattern);
    propagate();
    for (i=0; i < nrofoutputs; i++) {
      outputunit->get(i)->calc_connec_error();
      *pattern_error += fabs(outputunit->get(i)->get_unit_error());
    }
    *pattern_error /= nrofoutputs;
  }
  void trainoutput_set(Pattern_setC* trainingset, double *set_error) {
    int i,k;
    double learning_rate;
    double pattern_error;
    *set_error = 0.0;
    for (k=0; k < trainingset->nrofpatterns; k++) {
      trainoutput_pattern(trainingset->pattern[k], &pattern_error);
      *set_error += pattern_error;
    }
    learning_rate = (double)  0.175 / trainingset->nrofpatterns;
    for (i=0; i < nrofoutputs; i++)
      outputunit->get(i)->change_weights(learning_rate);
    *set_error /= trainingset->nrofpatterns;
    //cout << "de som van fouten over alle patronen en outputknopen: "
    //     << error << endl;
    // batch-training
  }
  void traininput_pattern(PatternC* trainingpattern) {
    int i;
    clamp(trainingpattern);
    propagate();
    for (i=0; i < nrofoutputs; i++)
      outputunit->get(i)->calc_unit_error();
  }
  void traininput_set(Pattern_setC* trainingset, double *best_candidate_value, int *best_candidate_index) {
    int k, i, c;
    double learning_rate;
    double cand_aver[nrofcandidates];
    double outputerror_aver[nrofoutputs];
    double sign_covar[nrofcandidates][nrofoutputs];
    double covariance[nrofcandidates]; // S
    double sumsqerror[nrofcandidates];
    // init
    for (c=0;c<nrofcandidates;c++) {
      covariance[c] = 0.0;
      cand_aver[c] = 0.0;
      sumsqerror[c] = 0.0;
      for (i=0; i<nrofoutputs; i++)
        sign_covar[c][i] = 0.0;
    }
    for (i=0; i<nrofoutputs; i++)
      outputerror_aver[i] = 0.0;
    // gemiddeldes berekenen:
    for (k=0; k < trainingset->nrofpatterns; k++) {
      traininput_pattern(trainingset->pattern[k]);
      for (c=0; c < nrofcandidates; c++)
        cand_aver[c] += candidateunit->get(c)->get_activation();
      for (i=0; i < nrofoutputs; i++)
        outputerror_aver[i] += outputunit->get(i)->get_unit_error();
    }   
    for (c=0; c < nrofcandidates; c++)
      cand_aver[c] /= trainingset->nrofpatterns;
    for (i=0; i < nrofoutputs; i++)
      outputerror_aver[i] /= trainingset->nrofpatterns;
    // signs van covarianties bereken en hoogste covariantie berekenen:
    for (k=0; k < trainingset->nrofpatterns; k++) {
      traininput_pattern(trainingset->pattern[k]);
      for (c=0; c < nrofcandidates; c++)
        for (i=0; i < nrofoutputs; i++) {
          sign_covar[c][i] += (candidateunit->get(c)->get_activation() - cand_aver[c])*(outputunit->get(i)->get_unit_error() - outputerror_aver[i]);
          covariance[c] += fabs(sign_covar[c][i]);
          sumsqerror[c] += pow((outputunit->get(i)->get_unit_error() - outputerror_aver[i]), 2); // anders dan cascor1.c
        }
    }
    for (c=0; c < nrofcandidates; c++) {
      //cout << "covariance = " << covariance[c] << "  sumsqerror = " << sumsqerror[c] << endl;
      covariance[c] /= sumsqerror[c];
      if (covariance[c] > *best_candidate_value) {
        *best_candidate_value = covariance[c];
        *best_candidate_index = c;
      }
      for (i=0; i < nrofoutputs; i++) {
        if (sign_covar[c][i] > 0)
          sign_covar[c][i] = +1;
        else
          sign_covar[c][i] = -1;
      }
    }
    // eindelijk dan dus de fout berekenen:
    for (k=0; k < trainingset->nrofpatterns; k++) {
      traininput_pattern(trainingset->pattern[k]);
      for (c=0; c < nrofcandidates; c++)
        for (i=0; i < nrofoutputs; i++) {
          // candidateunit->get(c)->add_unit_error(sign_covar[c][i] * (outputunit->get(i)->get_unit_error() - outputerror_aver[i]) * candidateunit->get(c)->get_activation() * (1 - candidateunit->get(c)->get_activation())); // voor asymetrische sigmoide
          candidateunit->get(c)->add_unit_error(sign_covar[c][i] * (outputunit->get(i)->get_unit_error() - outputerror_aver[i]) * (0.25 - pow(candidateunit->get(c)->get_activation(), 2)));
        }
    }
    learning_rate = (double) 1.0 / (trainingset->nrofpatterns * (nrofinputs + nrofhiddens));
    for (c=0; c < nrofcandidates; c++)
      candidateunit->get(c)->change_weights(learning_rate);
  }
  int test_pattern(PatternC* testpattern) {
    int i;
    int  error = 0;
    clamp(testpattern);
    propagate();
    if (outputunit->get(0)->calc_test_error() || outputunit->get(0)->calc_test_error())
      error = 1;
    return (error);
  }
};

int main(void)
{
  srand(time(0));
  int i, nepfout = 99;

  //char filename[20];
  Balance_scale_trainingsetC pats;  
  Balance_scale_testsetC testpats;
  Balance_scale_testset_newC all_testpats;
  pats.allprint();
  fstream outputfile("BalanceLog.data", ios::app);
  fstream outputfile3("blogb1.data", ios::app);
  Cascade_netC net1(pats.nrofinputs, pats.nroftargets);
  cout << "de testset-fout is:" << net1.test_set(&testpats) << endl;
  cout << "de trainingset-fout is:" << net1.test_set(&pats) << endl;
  cout << "aantal epochs is:" << net1.nrofepochs << endl;
  cout << "aantal voorbeelden in traininsset:" << pats.nrofpatterns << endl;
  
  net1.output_phase(&pats, &testpats, &all_testpats);
  cout << "de testset-fout is:" << net1.test_set(&testpats) << endl;
  cout << "de trainingset-fout is:" << net1.test_set(&pats) << endl;
  cout << "aantal epochs is:" << net1.nrofepochs << endl;
  cout << "aantal voorbeelden in traininsset:" << pats.nrofpatterns << endl;
  for (i=0; (nepfout > 0) && (net1.nrofepochs < 300); i++) {
    net1.infoprint();
    //sprintf(filename, "output%d.data", i); 
    //net1.test_net(filename);
    net1.input_phase(&pats);
    outputfile3 << net1.nrofepochs << endl;
    net1.output_phase(&pats, &testpats, &all_testpats);
    nepfout = net1.test_set(&testpats);
    cout << "de testset-fout is:" << nepfout << endl;
    cout << "de trainingset-fout is:" << net1.test_set(&pats) << endl;
    cout << "aantal epochs is:" << net1.nrofepochs << endl;
    cout << "aantal voorbeelden in traininsset:" << pats.nrofpatterns << endl;
  }
  net1.infoprint();
  //net1.test_net("outputlaatste.data");
  outputfile << " aantal epochs: " << net1.nrofepochs << "  aantal hiddens: " << i << endl;
  
  return(0);
};
