/*
// cascade.cc
//
*/
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <ctime>
#include "weightlistSP.cc"
#include "pattern.cc" 

using namespace std;

class Neural_networkC {};

class Cascade_netC : public Neural_networkC {
public:
  int nrofepochs;
  Cascade_netC (int nroi, int nroo) {
    int i, j;
    nrofinputs = nroi;
    nrofoutputs = nroo;
    nrofhiddens = 0;
    nrofcandidates = 0;
    nrofepochs = 0;
    inputunit = new Unit_array(nrofinputs);
    for(j=0;j<nrofinputs;j++)
      inputunit->set(j, new InputUnitC());
    outputunit = new Unit_array(nrofoutputs);
    for(i=0;i<nrofoutputs;i++)
      outputunit->set(i, new OutputUnitC());
    hiddenunit = new Unit_array(nrofhiddens);
    for(i=0; i < nrofoutputs; i++)
      for(j=0; j < nrofinputs; j++)
        outputunit->get(i)->add_connection(inputunit->get(j));
    candidateunit = NULL;
  }
  ~Cascade_netC() {
    int i, j;
    for(j=0;j<nrofinputs;j++)
      delete inputunit->get(j);
    delete inputunit;
    for(i=0;i<nrofoutputs;i++)
      delete outputunit->get(i);
    delete outputunit;
    for(i=0;i<nrofhiddens;i++)
      delete hiddenunit->get(i);
    delete hiddenunit;
    if (candidateunit) {
      for(i=0;i<nrofcandidates;i++)
        delete candidateunit->get(i);
      delete candidateunit;
    }
  }
  void infoprint() {
    int i;
    cout << "netwerk-inf: " 
    //     << "Aantal input-knopen: " << nrofinputs << endl
    //     << "Aantal output-knopen: " << nrofoutputs << endl
         << "Aantal hidden-knopen: " << nrofhiddens << endl;
    //     << "Aantal kandidaat-knopen: " << nrofcandidates << endl;
    //cout << "connecties van output-knopen: " << endl;
    for(i=0; i<nrofhiddens; i++)
      hiddenunit->get(i)->show();
    for(i=0; i<nrofoutputs; i++)
      outputunit->get(i)->show();
  }
  void output_phase(Pattern_setC* patternset) {
    double error = 1.0,
           old_error = 0.0;
    int output_patience;
    const int nrofe_atbegin = nrofepochs; 
    for (output_patience=0; (output_patience < 8) && ((nrofepochs - nrofe_atbegin) < 200); output_patience++) {
      //infoprint();
      nrofepochs++;
      trainoutput_set(patternset, &error);
      if (fabs(error - old_error) > (old_error * 0.005)) {
        output_patience = 0;
        old_error = error;
      }
      if (nrofepochs % 10 == 0)
        cout << " de echte fout is :" << error << endl;
    }
  }
  void input_phase(Pattern_setC* patternset) {
    int input_patience;
    double best_candidate_value,
           old_best_candidate_value = 0.0;
    int best_candidate_index = -1;
    const int nrofe_atbegin = nrofepochs; 
    newcandidates();
    for (input_patience=0; (input_patience < 30) && ((nrofepochs - nrofe_atbegin) < 200); input_patience++) {
      //infoprint();
      nrofepochs++;
      traininput_set(patternset, &best_candidate_value, &best_candidate_index);
      if (fabs(best_candidate_value - old_best_candidate_value) > (old_best_candidate_value * 0.03)) {
        input_patience = 0;
        old_best_candidate_value = best_candidate_value;
      }
      if (nrofepochs % 10 == 0)
        cout << "input_phase: de hoogste waarde is: " << best_candidate_value << " van kandidaat: " << best_candidate_index << endl;
    }
    movecandidate(best_candidate_index);
  }
  int test_set(Pattern_setC* testingset) {
    int k;
    int error = 0;
    for (k=0; k < testingset->nrofpatterns; k++)
      error += test_pattern(testingset->pattern[k]);
    //error /= testingset->nrofpatterns;
    //cout << "de som van fouten over alle patronen en outputknopen: "
    //     << error << endl;
    return error;
  }
  void test_net(const char* filename) {
    double x,y;
    ofstream evaluationdata (filename);
    inputunit->get(0)->clamp(1.0);
    for (x=-7.0;x<7.0; x += 0.01)
      for (y=-7.0; y < 7.0; y += 0.01) {
        inputunit->get(1)->clamp(x);
        inputunit->get(2)->clamp(y);
        propagate();
        evaluationdata << x << "  " << y << "  " << outputunit->get(0)->get_activation() << endl;        
      }
    evaluationdata.close();
  }
private: 
  int nrofinputs,
      nrofoutputs,
      nrofhiddens,
      nrofcandidates;
  Unit_array *inputunit,
             *outputunit,
             *hiddenunit,
             *candidateunit;
  void newcandidates() {
    int i,j;
    nrofcandidates = 8;
    candidateunit = new Unit_array(nrofcandidates);
    for (i=0; i < nrofcandidates; i++) {
      candidateunit->set(i, new HiddenUnitC());
      for(j=0; j < nrofinputs; j++)
        candidateunit->get(i)->add_connection(inputunit->get(j));
      for(j=0; j < nrofhiddens; j++)
        candidateunit->get(i)->add_connection(hiddenunit->get(j));
    }
  }
  void movecandidate(int winnercandidate) {
    int i,j;
    hiddenunit->add();
    nrofhiddens++;
    hiddenunit->set(nrofhiddens-1, candidateunit->get(winnercandidate));
    for(i=0; i<nrofoutputs; i++)
      outputunit->get(i)->add_connection(hiddenunit->get(nrofhiddens-1));
    for(i=0;i<nrofcandidates;i++)
      if (i != winnercandidate)
        delete candidateunit->get(i);
    delete candidateunit;
    candidateunit = NULL;
    nrofcandidates = 0;
  }
  void clamp(PatternC* pattern) {
    int i,j;
    for (j=0; j < nrofinputs; j++)
      inputunit->get(j)->clamp(pattern->input[j]);
    for (i=0; i < nrofoutputs; i++)
      outputunit->get(i)->set_target(pattern->target[i]);
  } 
  void propagate() {
    int i;
    for (i=0; i < nrofhiddens; i++)
      hiddenunit->get(i)->propagate();  // hier is de volgorde van belang
    for (i=0; i < nrofoutputs; i++)
      outputunit->get(i)->propagate();
    for (i=0; i < nrofcandidates; i++)
      candidateunit->get(i)->propagate();
  }
  void trainoutput_pattern(PatternC* trainingpattern, double *pattern_error) {
    int i;
    *pattern_error = 0.0;
    clamp(trainingpattern);
    propagate();
    for (i=0; i < nrofoutputs; i++) {
      outputunit->get(i)->calc_connec_error();
      *pattern_error += fabs(outputunit->get(i)->get_unit_error());
    }
    *pattern_error /= nrofoutputs;
  }
  void trainoutput_set(Pattern_setC* trainingset, double *set_error) {
    int i,k;
    double learning_rate;
    double pattern_error;
    *set_error = 0.0;
    for (k=0; k < trainingset->nrofpatterns; k++) {
      trainoutput_pattern(trainingset->pattern[k], &pattern_error);
      *set_error += pattern_error;
    }
    learning_rate = (double)  0.1 / trainingset->nrofpatterns;
    for (i=0; i < nrofoutputs; i++)
      outputunit->get(i)->change_weights(learning_rate);
    *set_error /= trainingset->nrofpatterns;
    //cout << "de som van fouten over alle patronen en outputknopen: "
    //     << error << endl;
    // batch-training
  }
  void traininput_pattern(PatternC* trainingpattern) {
    int i;
    clamp(trainingpattern);
    propagate();
    for (i=0; i < nrofoutputs; i++)
      outputunit->get(i)->calc_unit_error();
  }
  void traininput_set(Pattern_setC* trainingset, double *best_candidate_value, int *best_candidate_index) {
    int k, i, c;
    double learning_rate;
    double cand_aver[nrofcandidates];
    double outputerror_aver[nrofoutputs];
    double sign_covar[nrofcandidates][nrofoutputs];
    double covariance[nrofcandidates]; // S
    double sumsqerror[nrofcandidates];
    // init
    for (c=0;c<nrofcandidates;c++) {
      covariance[c] = 0.0;
      cand_aver[c] = 0.0;
      sumsqerror[c] = 0.0;
      for (i=0; i<nrofoutputs; i++)
        sign_covar[c][i] = 0.0;
    }
    for (i=0; i<nrofoutputs; i++)
      outputerror_aver[i] = 0.0;
    // gemiddeldes berekenen:
    for (k=0; k < trainingset->nrofpatterns; k++) {
      traininput_pattern(trainingset->pattern[k]);
      for (c=0; c < nrofcandidates; c++)
        cand_aver[c] += candidateunit->get(c)->get_activation();
      for (i=0; i < nrofoutputs; i++)
        outputerror_aver[i] += outputunit->get(i)->get_unit_error();
    }   
    for (c=0; c < nrofcandidates; c++)
      cand_aver[c] /= trainingset->nrofpatterns;
    for (i=0; i < nrofoutputs; i++)
      outputerror_aver[i] /= trainingset->nrofpatterns;
    // signs van covarianties bereken en hoogste covariantie berekenen:
    for (k=0; k < trainingset->nrofpatterns; k++) {
      traininput_pattern(trainingset->pattern[k]);
      for (c=0; c < nrofcandidates; c++)
        for (i=0; i < nrofoutputs; i++) {
          sign_covar[c][i] += (candidateunit->get(c)->get_activation() - cand_aver[c])*(outputunit->get(i)->get_unit_error() - outputerror_aver[i]);
          covariance[c] += fabs(sign_covar[c][i]);
          sumsqerror[c] += pow((outputunit->get(i)->get_unit_error() - outputerror_aver[i]), 2); // anders dan cascor1.c
        }
    }
    for (c=0; c < nrofcandidates; c++) {
      //cout << "covariance = " << covariance[c] << "  sumsqerror = " << sumsqerror[c] << endl;
      covariance[c] /= sumsqerror[c];
      if (covariance[c] > *best_candidate_value) {
        *best_candidate_value = covariance[c];
        *best_candidate_index = c;
      }
      for (i=0; i < nrofoutputs; i++) {
        if (sign_covar[c][i] > 0)
          sign_covar[c][i] = +1;
        else
          sign_covar[c][i] = -1;
      }
    }
    // eindelijk dan dus de fout berekenen:
    for (k=0; k < trainingset->nrofpatterns; k++) {
      traininput_pattern(trainingset->pattern[k]);
      for (c=0; c < nrofcandidates; c++)
        for (i=0; i < nrofoutputs; i++) {
          // candidateunit->get(c)->add_unit_error(sign_covar[c][i] * (outputunit->get(i)->get_unit_error() - outputerror_aver[i]) * candidateunit->get(c)->get_activation() * (1 - candidateunit->get(c)->get_activation())); // voor asymetrische sigmoide
          candidateunit->get(c)->add_unit_error(sign_covar[c][i] * (outputunit->get(i)->get_unit_error() - outputerror_aver[i]) * (0.25 - pow(candidateunit->get(c)->get_activation(), 2)));
        }
    }
    learning_rate = (double) 0.01 / (nrofinputs + nrofhiddens); // ook wel 1.92 / (trainingset->nrofpatterns * (nrofinputs + nrofhiddens))
    for (c=0; c < nrofcandidates; c++)
      candidateunit->get(c)->change_weights(learning_rate);
  }
  int test_pattern(PatternC* testpattern) {
    int i;
    int  error = 0;
    clamp(testpattern);
    propagate();
    for (i=0; i < nrofoutputs; i++) {
      error += outputunit->get(i)->calc_test_error();
    }
    return (error);
  }
};

int main(void)
{
  srand(time(0));
  int i, nepfout = 99;
  Pattern_setC pats2;
  char filename[20];
  //Balance_scale_trainingsetC pats;  
  pats2.read("spiraal.data");
  pats2.allprint();
  fstream outputfile("SpiralLog.data", ios::app);
  Cascade_netC net1(pats2.nrofinputs, pats2.nroftargets);
  
  net1.output_phase(&pats2);
  cout << "de nep-fout is:" << net1.test_set(&pats2) << endl;
  for (i=0; nepfout > 0; i++) {
    net1.infoprint();
    sprintf(filename, "output%d.data", i); 
    //net1.test_net(filename);
    net1.input_phase(&pats2);
    net1.output_phase(&pats2);
    nepfout = net1.test_set(&pats2);
    cout << "de nep-fout is:" << nepfout << endl;
    cout << "aantal epochs is:" << net1.nrofepochs << endl;
  }
  net1.infoprint();
  //net1.test_net("outputlaatste.data");
  outputfile << " aantal epochs: " << net1.nrofepochs << "  aantal hiddens: " << i << endl;
  
  return(0);
};
/* te doen:
 *  ok- niet gewone sigmoid gebruiken maar -1/2 zoals ook in cascor1.c
 * stop-criteria proberen na te bootsen
 *  ok- weight-decay zoeken in cascor1.c en toevoegen
 *  ok-beste kandidaat zoeken
 * verschillende quickproppen voor input en output zoals cascor1.c misschien
 * spiral.data plotten, kijken of ok +uitzoeken waaarom cascade het niet lukt
 * meer....
 */
