/*
// pattern.cc
// Olaf Booij Olaf
// last modified: 25-7-2002
//
*/
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <ctime>

using namespace std;

class PatternC {
public:
  double *input;
  double *target;
  PatternC (int nrofinputs, int nroftargets) {
    input = new double[nrofinputs];
    target = new double[nroftargets];
  }
  ~PatternC () {
    delete [] input;
    delete [] target;
  }
  void allprint(int nrofinputs, int nroftargets) {
    int i;
    for(i=0; i < nrofinputs; i++)
      cout << " " << input[i] << "   ";
    cout << endl;
    for(i=0; i < nroftargets; i++)
      cout << " " << target[i] << "   ";
    cout << endl; 
  }
  void read(ifstream *streamname, int nrofinputs, int nroftargets) {
    int i;
    for(i=0; i < nrofinputs; i++)
      *streamname >> input[i];
    for(i=0; i < nroftargets; i++)
      *streamname >> target[i];
  }
  void copyto(PatternC *pattern_copy, int nrofinputs, int nroftargets) {
    int i;
    for (i=0;i<nrofinputs;i++)
      pattern_copy->input[i] = input[i];
    for (i=0;i<nroftargets;i++)
      pattern_copy->target[i] = target[i];
  }
};

class Pattern_setC {
public:
  int nrofpatterns,
      nrofinputs,
      nroftargets;
  PatternC **pattern;
  Pattern_setC () {
    nrofpatterns = -1;
    pattern = NULL;
  }
  Pattern_setC(int nrop, int nroi, int nrot) {
    int i;
    nrofpatterns = nrop;
    nrofinputs = nroi;
    nroftargets = nrot;
    pattern = new PatternC*[nrofpatterns];
    for (i=0; i < nrofpatterns; i++)
      pattern[i] = new PatternC(nrofinputs, nroftargets);
  }
  virtual ~Pattern_setC () {
    if (pattern) delete [] pattern;
  }
  virtual void expand() {show_error();}
  void allprint() {
    int i;
    cout << "Number of patterns: " << nrofpatterns << endl
         << "Aantal input-knopen: " << nrofinputs
         << "; aantal output-knopen: " << nroftargets << endl;
      for (i=0; i < nrofpatterns; i++)
        pattern[i]->allprint(nrofinputs, nroftargets);
  }
  void read(const char* filename) {
    int i;
    if (pattern) 
      cerr << "Pattern_setC: error: pattern already instantiated" << endl;
    ifstream patternstream (filename);
    patternstream >> nrofpatterns;
    patternstream >> nrofinputs;
    patternstream >> nroftargets;
    pattern = new PatternC*[nrofpatterns];
    for (i=0; i < nrofpatterns; i++) {
      pattern[i] = new PatternC(nrofinputs, nroftargets);
      pattern[i]->read(&patternstream, nrofinputs, nroftargets);
    }
    patternstream.close();
  }
  void kill() { // not in the standard destructor, because they are used when copying (maybe ugly)
    int i;
    for (i=0; i < nrofpatterns; i++)
      delete pattern[i];
  }
  void copy(int pattern_index, PatternC *target_pattern) {
    if ((pattern == NULL) || (target_pattern == NULL))
      cerr << "Pattern_setC: error: illegal copying" << endl;
    else
      pattern[pattern_index]->copyto(target_pattern, nrofinputs, nroftargets);
  }
private:
  void show_error() {
    cout << "error in Pattern_setC: function-call-error" << endl;
  }
};

class Balance_scale_problemsC : public Pattern_setC {
public:
  Balance_scale_problemsC(int nrop) : Pattern_setC(nrop, 5, 2) {}
//protected:
  void get_all_BSPs() {
    double wl,
           dl,
           wr,
           dr;
    int i = 0;
    for (wl=1;wl<=5;wl++)
      for (dl=1;dl<=5;dl++)
        for (wr=1;wr<=5;wr++)
          for (dr=1;dr<=5;dr++) {
            pattern[i]->input[0] = 1.0; // bias
            pattern[i]->input[1] = wl;
            pattern[i]->input[2] = dl;
            pattern[i]->input[3] = wr;
            pattern[i]->input[4] = dr;
            if (wl*dl > wr*dr) {  // leftside will go down
              pattern[i]->target[0] = 0.5;
              pattern[i]->target[1] = -0.5;
            }  
            else if (wl*dl < wr*dr) { // rightside will go down
              pattern[i]->target[0] = -0.5;
              pattern[i]->target[1] = 0.5;
            }
            else if (wl*dl == wr*dr) { // scale is in balance
              pattern[i]->target[0] = 0.0;
              pattern[i]->target[1] = 0.0;
            } 
            else
              cerr << "Balance_scale_trainingsetC: hmmm.... something wrong with the balance" << endl;
            i++;
          }
  }
};

// trainingset maken voor balanstaak:
class Balance_scale_trainingsetC : public Balance_scale_problemsC {
public :
  Balance_scale_problemsC *BSP_all;
  Balance_scale_problemsC *BSP_equal_dist;
  Balance_scale_problemsC *BSP_rest;
  Balance_scale_trainingsetC() : Balance_scale_problemsC(401) { // 100 at the beginning + 300 for the epochs + 1 to be sure
    int i = 0,
        equal_dist_index = 0,
        rest_index = 0,
        rand_index;
    BSP_all = new Balance_scale_problemsC(625);
    BSP_equal_dist = new Balance_scale_problemsC(125);
    BSP_rest = new Balance_scale_problemsC(500);
    // dividing the problems in equal-distance-problems and not-equal-distance-problems:
    // this is used for the biasing in favor of equal distance
    BSP_all->get_all_BSPs();
    for(i=0; i<BSP_all->nrofpatterns ; i++) {
      if (BSP_all->pattern[i]->input[2] == BSP_all->pattern[i]->input[4]) {  // equal dist
        BSP_equal_dist->pattern[equal_dist_index] = BSP_all->pattern[i];
        equal_dist_index++;
      }
      else  { // rest
        BSP_rest->pattern[rest_index] = BSP_all->pattern[i];
        rest_index++;
      }
    }
    delete BSP_all; 
    // now pick a 100 random balance scale problems without replacement 
    // with a 0.9 bias in favor of equal distance: 
    nrofpatterns = 100;
    for (i=0; i<nrofpatterns; i++) {
      if (rand()%10 < 9) {  // chance = 90%
        rand_index = rand() % BSP_equal_dist->nrofpatterns;
        pattern[i] = BSP_equal_dist->pattern[rand_index]; 
        BSP_equal_dist->nrofpatterns--;
        BSP_equal_dist->pattern[rand_index] = BSP_equal_dist->pattern[BSP_equal_dist->nrofpatterns];
      }
      else {
        rand_index = rand() % BSP_rest->nrofpatterns;
        pattern[i] = BSP_rest->pattern[rand_index]; 
        BSP_rest->nrofpatterns--;
        BSP_rest->pattern[rand_index] = BSP_rest->pattern[BSP_rest->nrofpatterns];
      }
    }
  }
  ~Balance_scale_trainingsetC() {
    nrofpatterns = 100;
    kill();
    BSP_equal_dist->kill();
    BSP_rest->kill();
    delete BSP_equal_dist;
    delete BSP_rest;
  }
  void expand() {
    int rand_index;
    // expand the trainingset by adding a random balance scale problems with replacement 
    // with a 0.9 bias in favor of equal distance: 
    if (rand()%10 < 9) {  // chance = 90%
      rand_index = rand() % BSP_equal_dist->nrofpatterns;
      pattern[nrofpatterns] = BSP_equal_dist->pattern[rand_index]; 
    }
   else {
      rand_index = rand() % BSP_rest->nrofpatterns;
      pattern[nrofpatterns] = BSP_rest->pattern[rand_index]; 
    }
    nrofpatterns++;
    if (nrofpatterns == 401)
      cerr << "error: Balance_scale_trainingsetC: number of patterns is too large (>400)" << endl;
  }
};

// Balance_scale_testsetC
// very ugly, but doesn't produce memory-leaks or dangling pointers (i hope)
class Balance_scale_testsetC : public Balance_scale_problemsC {
public :
  Balance_scale_problemsC *BSP_all;
  Balance_scale_problemsC ***BSP_problemtypes;           // equal weight, equal distance
  enum problemtypeE {balance, weight, distance, conflict_weight, conflict_distance, conflict_balance, no_problem};
  enum torque_difference_typeE {torque0, torque1, torque2to5, torque6to9, torque10to20};
  Balance_scale_testsetC() : Balance_scale_problemsC(24) {
    int i,
        j,
        j2,  // creativity is gone
        problemtypes_index[6][5],
        rand_index,
        problem_index = 0,
        torque_difference;
    enum problemtypeE problemtype; 
    enum torque_difference_typeE torque_difference_type;
    BSP_all = new Balance_scale_problemsC(625);
    BSP_problemtypes = new Balance_scale_problemsC**[6];
    for (i=0;i<6;i++) {
      BSP_problemtypes[i] = new Balance_scale_problemsC*[5];
      for (j=0;j<5;j++) {
        BSP_problemtypes[i][j] = new Balance_scale_problemsC(400);
        problemtypes_index[i][j] = 0;
      }
    }
    // dividing the problems into groups of different problems and torgue-differences 
    BSP_all->get_all_BSPs();
    for(i=0; i<BSP_all->nrofpatterns ; i++) {
      if ((BSP_all->pattern[i]->input[2] == BSP_all->pattern[i]->input[4]) && (BSP_all->pattern[i]->input[1] == BSP_all->pattern[i]->input[3]))
        problemtype = balance;
      else if (BSP_all->pattern[i]->input[2] == BSP_all->pattern[i]->input[4])
        problemtype = weight;
      else if (BSP_all->pattern[i]->input[1] == BSP_all->pattern[i]->input[3])
        problemtype = distance;
      else if (((BSP_all->pattern[i]->input[1] > BSP_all->pattern[i]->input[3]) && (BSP_all->pattern[i]->input[2] > BSP_all->pattern[i]->input[4])) || ((BSP_all->pattern[i]->input[1] < BSP_all->pattern[i]->input[3]) && (BSP_all->pattern[i]->input[2] < BSP_all->pattern[i]->input[4])))
        problemtype = no_problem;
      else if (((BSP_all->pattern[i]->input[1] > BSP_all->pattern[i]->input[3]) && (BSP_all->pattern[i]->target[0] == 0.5)) || ((BSP_all->pattern[i]->input[3] > BSP_all->pattern[i]->input[1]) && (BSP_all->pattern[i]->target[1] == 0.5)))
        problemtype = conflict_weight;
      else if (((BSP_all->pattern[i]->input[2] > BSP_all->pattern[i]->input[4]) && (BSP_all->pattern[i]->target[0] == 0.5)) || ((BSP_all->pattern[i]->input[4] > BSP_all->pattern[i]->input[2]) && (BSP_all->pattern[i]->target[1] == 0.5)))
        problemtype = conflict_distance;
      else if (((BSP_all->pattern[i]->input[1] != BSP_all->pattern[i]->input[3]) && (BSP_all->pattern[i]->input[2] != BSP_all->pattern[i]->input[4])) && (BSP_all->pattern[i]->target[0] == 0.0))
        problemtype = conflict_balance;
      else 
        cerr << "Balance_scale_testsetC: error: something wrong with the scale-problems" << endl;
      torque_difference = (int)fabs((BSP_all->pattern[i]->input[1] * BSP_all->pattern[i]->input[2]) - (BSP_all->pattern[i]->input[3] * BSP_all->pattern[i]->input[4]));
      if (torque_difference == 0)
        torque_difference_type = torque0; // this is for balance and conflict-balance 
      else if (torque_difference == 1)
        torque_difference_type = torque1;
      else if (torque_difference < 6)
        torque_difference_type = torque2to5;
      else if (torque_difference < 10)
        torque_difference_type = torque6to9;
      else torque_difference_type = torque10to20;
      if (problemtype != no_problem) {
        BSP_all->copy(i, BSP_problemtypes[problemtype][torque_difference_type]->pattern[problemtypes_index[problemtype][torque_difference_type]]);
        problemtypes_index[problemtype][torque_difference_type]++;
      }
    }
    BSP_all->kill();
    delete BSP_all; 
    // now pick a 24 random balance scale problems, 1 out of each group:
    for (i=0; i<6; i++)
      for (j=1; j<5; j++) {
        j2 = j;
        if ((i == 0) || (i == 5))  // balance has no torque 
          j2 = 0;
        cout << "problem-type: " << problemtypes_index[i][j2] << endl;
        rand_index = rand() % problemtypes_index[i][j2];
        BSP_problemtypes[i][j2]->copy(rand_index, pattern[problem_index]);
        problem_index++;
      }
    // now free some memory:
    for (i=0; i<6; i++) {
      for (j=0; j<5; j++) {
        BSP_problemtypes[i][j]->kill();
        delete BSP_problemtypes[i][j];
      } 
      delete [] BSP_problemtypes[i];
    }
    delete [] BSP_problemtypes;
  }
  ~Balance_scale_testsetC() {
    kill();
  }
};




class Balance_scale_testset_newC : public Balance_scale_problemsC {
public :
  Balance_scale_problemsC *BSP_all;
  Balance_scale_problemsC **BSP_problemtypes;           // equal weight, equal distance
  enum problemtypeE {balance, weight, distance, conflict_weight, conflict_distance, conflict_balance, no_problem};
  Balance_scale_testset_newC() : Balance_scale_problemsC(625) {
    int i,
        j,
        j2,  // creativity is gone
        problemtypes_index[6],
        rand_index,
        problem_index = 0,
        torque_difference;
    enum problemtypeE problemtype; 
    BSP_all = new Balance_scale_problemsC(625);
    BSP_problemtypes = new Balance_scale_problemsC*[6];
    for (i=0;i<6;i++) {
      BSP_problemtypes[i] = new Balance_scale_problemsC(200);
      problemtypes_index[i] = 0;
    }
    // dividing the problems into groups of different problems and torgue-differences 
    BSP_all->get_all_BSPs();
    for(i=0; i<BSP_all->nrofpatterns ; i++) {
      if ((BSP_all->pattern[i]->input[2] == BSP_all->pattern[i]->input[4]) && (BSP_all->pattern[i]->input[1] == BSP_all->pattern[i]->input[3]))
        problemtype = balance;
      else if (BSP_all->pattern[i]->input[2] == BSP_all->pattern[i]->input[4])
        problemtype = weight;
      else if (BSP_all->pattern[i]->input[1] == BSP_all->pattern[i]->input[3])
        problemtype = distance;
      else if (((BSP_all->pattern[i]->input[1] > BSP_all->pattern[i]->input[3]) && (BSP_all->pattern[i]->input[2] > BSP_all->pattern[i]->input[4])) || ((BSP_all->pattern[i]->input[1] < BSP_all->pattern[i]->input[3]) && (BSP_all->pattern[i]->input[2] < BSP_all->pattern[i]->input[4])))
        problemtype = no_problem;
      else if (((BSP_all->pattern[i]->input[1] > BSP_all->pattern[i]->input[3]) && (BSP_all->pattern[i]->target[0] == 0.5)) || ((BSP_all->pattern[i]->input[3] > BSP_all->pattern[i]->input[1]) && (BSP_all->pattern[i]->target[1] == 0.5)))
        problemtype = conflict_weight;
      else if (((BSP_all->pattern[i]->input[2] > BSP_all->pattern[i]->input[4]) && (BSP_all->pattern[i]->target[0] == 0.5)) || ((BSP_all->pattern[i]->input[4] > BSP_all->pattern[i]->input[2]) && (BSP_all->pattern[i]->target[1] == 0.5)))
        problemtype = conflict_distance;
      else if (((BSP_all->pattern[i]->input[1] != BSP_all->pattern[i]->input[3]) && (BSP_all->pattern[i]->input[2] != BSP_all->pattern[i]->input[4])) && (BSP_all->pattern[i]->target[0] == 0.0))
        problemtype = conflict_balance;
      else 
        cerr << "Balance_scale_testsetC: error: something wrong with the scale-problems" << endl;
      if (problemtype != no_problem) {
        BSP_all->copy(i, BSP_problemtypes[problemtype]->pattern[problemtypes_index[problemtype]]);
        problemtypes_index[problemtype]++;
      }
      else
        problemtypes_index[problemtype]++;
    }
    BSP_all->kill();
    delete BSP_all; 
    // now pick a 24 random balance scale problems, 1 out of each group:
    for (i=0; i<6; i++) {
      cout << "problem-type: " << problemtypes_index[i] << endl;
      for (rand_index = 0; rand_index < problemtypes_index[i]; rand_index++) {
        BSP_problemtypes[i]->copy(rand_index, pattern[problem_index]);
        problem_index++;
      }
    }
    // now free some memory:
    for (i=0; i<6; i++) {
      delete BSP_problemtypes[i];
    }
    delete [] BSP_problemtypes;
  }
  ~Balance_scale_testset_newC() {
    kill();
  }
};

/*
int main(void) {
  srand(time(0));
  Balance_scale_trainingsetC balproblemen;
  Balance_scale_testsetC balproblemen2;
//  balproblemen.allprint();
//  balproblemen.expand();
//  balproblemen.allprint();
  balproblemen2.allprint();
  return(0);
}
*/
