/*
// unitarray.cc
// Olaf Booij
// last modified: 27-6-2002
//
*/
/*
// weightlist.cc
// Olaf Booij
// last modified: 21-7-2002
//
*/
#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <climits>
#include <ctime>
#include <cmath>

using namespace std;

class UnitC;

//  weights: wij j->i (dus i=output, j=input)
class ConnectionC {
public:
  ConnectionC(UnitC* output_u, UnitC* input_u) {
    // voor output: learning_rate = 0.35 / nrofpatterns
    // voor input : learning_rate = 1.0 / (nrofpatterns*(nrofinputs+nrofhiddens)
    // momentum = 0.3; // zeer natte vinger...
    // maximum_growth_factor = 1.75;
    input_unit = input_u;
    output_unit = output_u;
    old_weight_error = 0.0;
    old_weight_change = 0.0;
    weight_error = 0.0;
    // gewicht tussen -1.0 en +1.0:
    weight = (double) (rand() % INT_MAX) / INT_MAX * 2.0 - 1.0;
  }
  //~ConnectionC() {}
  void show() {
    cout << weight;
  }
  double get_input(); // pas gedefinieerd na unitC-defenitie
  void add_weight_error(double output_error); // pas gedef na UnitC
  int in_output_phase();
  void change_weight(double learning_rate) {
    // quickprop:
    double weight_change = 0.0;
    if (in_output_phase())
      weight_error -= 0.0001 * weight; // hier zat die rotzak: weight_error = -slope
    if (old_weight_change < 0.0) {
      if (weight_error < 0.0)
        weight_change += learning_rate * weight_error;
      if (weight_error <= 0.6667 * old_weight_error)
        weight_change += 2.0 * old_weight_change;
      else
        weight_change += old_weight_change * weight_error/(old_weight_error - weight_error);
    }
    else if (old_weight_change > 0.0) {
      if (weight_error > 0.0)
        weight_change += learning_rate * weight_error;
      if (weight_error >= 0.6667 * old_weight_error)
        weight_change += 2.0 * old_weight_change;
      else
        weight_change += old_weight_change * weight_error/(old_weight_error - weight_error);
    }
    else
      weight_change += learning_rate * weight_error;
/*      weight_change = 0.0;
    else
      weight_change = weight_error/(old_weight_error - weight_error) * old_weight_change;
    if (fabs(weight_change) > maximum_growth_factor * fabs(old_weight_change))
        weight_change = maximum_growth_factor * old_weight_change;
    if (weight_error * old_weight_error <= 0.0)  // the signs of the slopes are equal
      weight_change += learning_rate * weight_error;
    // weight_change = learning_rate * weight_error + momentum * old_weight_change; // backprop
    //  cout << weight_error << endl;
*/
    weight = weight + weight_change;
    old_weight_change = weight_change;
    old_weight_error = weight_error; // voor cascor
    weight_error = 0.0;
  }
private:
  UnitC *input_unit,
        *output_unit; // hoeft niet echt bij cascor
  double learning_rate,
         // momentum, // voor backprop
         maximum_growth_factor;
  double weight,
         old_weight_change,
         weight_error,
         old_weight_error;
};

// overgeschreven van "on to c++"
// doubles in een lijst
class ConnecLinkC {
public:
  ConnecLinkC *next;
  ConnectionC *element;
  ConnecLinkC(UnitC *output, UnitC *input, ConnecLinkC *l) {
    element = new ConnectionC(output, input);
    next = l;
  }
  ~ConnecLinkC() {
    delete element;
  }
};

class ConnecListC {
public:
  ConnecListC() {
    top = NULL;
    current = top;
  }
  ~ConnecListC() {
    while(top)
      remove();
  } 
  void add(UnitC* output, UnitC* input) { 
    top = new ConnecLinkC(output, input, top);
    current = top; // is dit fijn??
  }
  void remove() {
    ConnecLinkC *previous = top;
    top = top->next;
    delete previous;
  }
  void advance() {
    current= current->next;
  }
  ConnectionC* get() {
    return current->element;
  }
  int eol() {
    return ! current;
  }
  void reset() {
    current = top;
  }
private:
  ConnecLinkC *top;
  ConnecLinkC *current;
};

// unit:
class UnitC {
public:
  UnitC () {activation = 0;}
  virtual ~UnitC() {}
  double get_activation() {return activation;}
  int get_type() {return type;}
  virtual void show() {show_error();}
  virtual void add_connection(UnitC *input_unit) {show_error();}
  virtual void clamp(double a) {show_error();}
  virtual void set_target(double t) {show_error();}
  virtual void change_weights(double learning_rate) {show_error();}
  virtual void calc_unit_error() {show_error();}
  virtual void calc_connec_error() {show_error();}
  virtual int calc_test_error() {show_error();}
  virtual void propagate() {show_error();}
  virtual double get_unit_error() {show_error();}
  virtual void add_unit_error(double error) {show_error();}
protected:
  enum typeT {INPUT_UNIT, HIDDEN_UNIT, OUTPUT_UNIT} type;
  void set_activation(double a) {activation = a;}
private:
  double activation;
  void show_error() {
    cout << "error in UnitC: function-call-error" << endl;
  }
};

// input-unit
class InputUnitC : public UnitC {
public:
  InputUnitC() : UnitC() {type = INPUT_UNIT;}
  ~InputUnitC() {}
  void clamp(double a) {set_activation(a);}
};

// niet-input-unit
class NonInputUnitC : public UnitC {
public:
  NonInputUnitC() : UnitC() {input_connections = new ConnecListC();}
  virtual ~NonInputUnitC() {
    delete input_connections;
  }
  void show() {
    for(input_connections->reset(); !input_connections->eol(); input_connections->advance()) {
      input_connections->get()->show();
      cout << "  ";
    }
    cout << endl;
  }
  void propagate() {
    double inputsum = 0.0;
    for(input_connections->reset(); !input_connections->eol(); input_connections->advance())
      inputsum += input_connections->get()->get_input();
    // threshold function:
    //if (inputsum > 0)
    //  set_activation(1);
    //else
    //  set_activation(-1);
    //  asymetric sigmoid function:
    //  set_activation(1 / (1 + exp(-inputsum)));
    // symetric sigmoid function between -0.5 en 0.5: 
    set_activation(1 / (1 + exp(-inputsum)) - 0.5);
    if (inputsum == 0) {
      cout << "error waarschijnlijk: inputsum = 0" << endl;
			input_connections->reset();
			if (input_connections->eol())
				cout << " er zijn ook helemaal geen inkomende connecties" << endl;
		}
  }
  void change_weights(double learning_rate) {
    for(input_connections->reset(); !input_connections->eol(); input_connections->advance()) 
      input_connections->get()->change_weight(learning_rate);
  }
  void add_connection(UnitC *input_unit) {
    input_connections->add(this, input_unit); // "this" = &UnitCobject
  }
protected:
  ConnecListC *input_connections;
};

// hidden-unit
class HiddenUnitC : public NonInputUnitC {
public:
  HiddenUnitC() : NonInputUnitC() {type = HIDDEN_UNIT;}
  ~HiddenUnitC() {}
  void add_unit_error(double error) {
    for(input_connections->reset(); !input_connections->eol(); input_connections->advance())
      input_connections->get()->add_weight_error(error);
  }
};

// ** er moet nog een candidate-unit zijn ofiets denkk.

// output-unit
class OutputUnitC : public NonInputUnitC {
public:
  OutputUnitC() : NonInputUnitC() {type = OUTPUT_UNIT;}
  ~OutputUnitC() {}
  void set_target(double t) {target = t;}
  double get_unit_error() {return unit_error;}
  void calc_unit_error() {
    unit_error = (target - get_activation()) * (0.1 + 0.25 - pow(get_activation(), 2)); // backprop met asymetrische sigmoid met de sigmoid_prime_offset van 0.1 
    // unit_error = (target - get_activation()) * get_activation() * (1 - get_activation()); // backprop met asymetrische sigmoid 
    //unit_error = target - get_activation(); // delta-regel
  }
  void calc_connec_error() {
    calc_unit_error();
    for(input_connections->reset(); !input_connections->eol(); input_connections->advance())
      input_connections->get()->add_weight_error(unit_error);
  }
  int calc_test_error() { // for testing the net's performence
    if (fabs(target - get_activation()) >= 0.25) {
      //cout << "fout!: target = " << target << " activation = " << get_activation() << endl;
      return(1);
    }
    else {
      //cout << "goed!: target = " << target << " activation = " << get_activation() << endl;
      return(0);
    }
  }
private:
  double target;
  double unit_error; // misschien moet dit ook voor hidden gelden
};

double ConnectionC::get_input()
{
  return (weight * input_unit->get_activation());
}
void ConnectionC::add_weight_error(double output_error)
{
  weight_error += output_error * input_unit->get_activation();
}
int ConnectionC::in_output_phase()
{
  if (output_unit->get_type() == 2)
    return(1);
  else
    return(0);
}

// opdracht 754 uit "on to c++"
class Unit_array {
public:
  Unit_array(int nrofe) {
    nr_of_elements = nrofe;
    hidden_array = new UnitC*[nr_of_elements];
  }
  ~Unit_array() {
    delete [] hidden_array;
  }
  UnitC* get(int n) {
    if (n < nr_of_elements) 
      return hidden_array[n];
    else { 
      cerr << "!unitarray: error: trying to get nonexisting element" << endl;
      return NULL;
    }
  }
  void set(int n, UnitC* unit) {
    if (n < nr_of_elements)
      hidden_array[n] = unit;
    else 
      cerr << "!unitarray: error: trying to get nonexisting element" << endl;
  }
  void add() {
    UnitC **new_hidden_array;
    int k;
    new_hidden_array = new UnitC*[nr_of_elements + 1];
    for (k=0; k < nr_of_elements; k++)
      new_hidden_array[k] = hidden_array[k];
    nr_of_elements++;
    delete [] hidden_array;
    hidden_array = new_hidden_array;
  }
private:
  int nr_of_elements;
  UnitC **hidden_array;
};

/*
int main(void)
{
  int i;
  srand(time(0));
  OutputUnitC eenoutputknoop;
  HiddenUnitC eenhiddenknoop,
              eenhiddenknoop2;
  InputUnitC eeninputknoop,
             eeninputknoop2;

  eenhiddenknoop.add_connection(&eeninputknoop);
  eenhiddenknoop.add_connection(&eeninputknoop2);
  eenhiddenknoop2.add_connection(&eeninputknoop);
  eenhiddenknoop2.add_connection(&eeninputknoop2);
  eenoutputknoop.add_connection(&eenhiddenknoop);
  eenoutputknoop.add_connection(&eenhiddenknoop2);
  cout << "activatie voor propagatie " << eenoutputknoop.get_activation() << endl;
  eeninputknoop.clamp(1.0);
  eeninputknoop2.clamp(1.0);
  eenoutputknoop.set_target(1.0);
  eenhiddenknoop.propagate();
  eenhiddenknoop2.propagate();
  for(i = 0; i<10; i++) {
    eenoutputknoop.propagate();
    eenoutputknoop.calc_error();
    eenoutputknoop.change_weights();
    cout << "activatie na propagatie " << eenoutputknoop.get_activation() << endl;
  }
  return(0); 
}*/
